from collections import namedtuple
from dataclasses import dataclass, field
from typing import Any, Dict, List, Optional, Set

from iam_ape.consts import PolicyElement
from iam_ape.helper_types import (
    ActionDict,
    AwsPolicyType,
    IneffectiveActionDict,
    PermissionsContainerDict,
)


class HashableList(List[Any]):
    def __hash__(self) -> int:  # type: ignore[override]
        return hash(tuple(self))


class HashableDict(Dict[Any, Any]):
    def __hash__(self) -> int:  # type: ignore[override]
        return hash(tuple(self.items()))

    @classmethod
    def recursively(cls, dict_obj: Optional[Dict[Any, Any]]):
        if dict_obj is None:
            return dict_obj
        for key, value in dict_obj.items():
            if isinstance(value, dict):
                dict_obj[key] = cls.recursively(value)
            elif isinstance(value, list):
                dict_obj[key] = HashableList(value)
            else:
                assert hasattr(value, "__hash__"), f"Unhashable type: {type(value)}"
                dict_obj[key] = value
        return cls(dict_obj)


@dataclass(unsafe_hash=True)
class Action:
    action: str
    resource: Optional[str]
    not_resource: Optional[str]
    condition: Optional[Dict[str, Any]]
    source: str

    def __post_init__(self) -> None:
        self.condition = HashableDict.recursively(self.condition)

    def to_dict(self) -> ActionDict:
        return {
            PolicyElement.ACTION: self.action,
            PolicyElement.RESOURCE: self.resource,
            PolicyElement.NOTRESOURCE: self.not_resource,
            PolicyElement.CONDITION: self.condition,
            "Source": self.source,
        }


@dataclass(unsafe_hash=True)
class IneffectiveAction(Action):
    denied_by: str

    def to_dict(self) -> IneffectiveActionDict:
        return {
            PolicyElement.ACTION: self.action,
            PolicyElement.RESOURCE: self.resource,
            PolicyElement.NOTRESOURCE: self.not_resource,
            PolicyElement.CONDITION: self.condition,
            "Source": self.source,
            "DeniedBy": self.denied_by,
        }


class PolicyWithSource(namedtuple("PolicyWithSource", ["source", "policy"])):
    source: str
    policy: AwsPolicyType


@dataclass
class PermissionsContainer:
    allowed_permissions: Dict[str, Set[Action]] = field(default_factory=dict)
    denied_permissions: Dict[str, Set[Action]] = field(default_factory=dict)
    ineffective_permissions: Set[IneffectiveAction] = field(default_factory=set)

    def to_dict(self) -> PermissionsContainerDict:
        res: PermissionsContainerDict = {
            "allowed_permissions": [],
            "denied_permissions": [],
            "ineffective_permissions": [],
        }
        for action_tuple_set in self.allowed_permissions.values():
            for action_tuple in action_tuple_set:
                res["allowed_permissions"].append(action_tuple.to_dict())
        for action_tuple_set in self.denied_permissions.values():
            for action_tuple in action_tuple_set:
                res["denied_permissions"].append(action_tuple.to_dict())
        for ineffective_action in self.ineffective_permissions:
            res["ineffective_permissions"].append(ineffective_action.to_dict())

        return res


@dataclass
class DenialEvalResult:
    should_deny: bool
    new_action_values: Set[Action]
    denied_by: Optional[Action]



import argparse
import json
import logging
import os
import re
import sys
from typing import Any, Dict, List, Optional, Tuple, Union

import boto3
from botocore.exceptions import ClientError, ProfileNotFound

from iam_ape.aws_iam_actions.scrape_iam_actions import scrape_iam_actions
from iam_ape.evaluator import AuthorizationDetails, EffectivePolicyEvaluator
from iam_ape.helper_classes import PolicyWithSource
from iam_ape.helper_functions import deep_update
from iam_ape.helper_types import AwsPolicyType, EntityType, FinalReportT

logger = logging.getLogger("IAM-APE")
logging.getLogger("botocore").setLevel(logging.WARNING)  # suppress botocore logs
entity_regex_string = r"arn:aws(-cn|-us-gov)?:iam::(?P<account>\d{12}):(?P<entity_type>user|group|role)/[\w-]+"
entity_regex = re.compile(entity_regex_string)

banner = """
  __   __   _  _        __   ____  ____ 
 (  ) / _\ ( \/ ) ___  / _\ (  _ \(  __)
  )( /    \/ \/ \(___)/    \ ) __/ ) _) 
 (__)\_/\_/\_)(_/     \_/\_/(__)  (____)
 By Orca Security  https://orca.security
"""


class LogFormatter(logging.Formatter):
    err_fmt = "ERROR: %(msg)s"
    dbg_fmt = "DBG: %(module)s: %(lineno)d: %(msg)s"
    info_fmt = "%(msg)s"

    def __init__(self):
        super().__init__(fmt="%(levelno)d: %(msg)s", datefmt=None, style="%")

    def format(self, record):
        format_orig = self._style._fmt

        if record.levelno == logging.DEBUG:
            self._style._fmt = LogFormatter.dbg_fmt

        elif record.levelno == logging.INFO:
            self._style._fmt = LogFormatter.info_fmt

        elif record.levelno == logging.ERROR:
            self._style._fmt = LogFormatter.err_fmt

        result = logging.Formatter.format(self, record)

        self._style._fmt = format_orig

        return result


def print_banner() -> None:
    logger.info("\033[93m" + banner + "\033[0m")


def initialize_logger() -> None:
    format = LogFormatter()
    handler = logging.StreamHandler(sys.stdout)
    handler.setFormatter(format)
    logging.root.addHandler(handler)
    logging.root.setLevel(logging.INFO)


def validate_arn(arn: str) -> Tuple[EntityType, str]:
    regex_match = entity_regex.fullmatch(arn)
    try:
        assert regex_match
    except AssertionError:
        raise ValueError(
            f'Invalid ARN format: "{arn}". Expected: "{entity_regex_string}"'
        )

    entity_type = regex_match.group("entity_type")
    account_id = regex_match.group("account")

    if entity_type == "user":
        return EntityType.user, account_id
    elif entity_type == "group":
        return EntityType.group, account_id
    else:
        return EntityType.role, account_id


def load_auth_details_from_json(inp: str) -> AuthorizationDetails:
    with open(inp, "r") as f:
        auth_report = json.load(f)
    return AuthorizationDetails(auth_report)


def load_auth_details_from_aws(profile: Optional[str] = None) -> AuthorizationDetails:
    logger.info("Attempting to fetch authorization details report from AWS...")
    auth_report: Dict[str, Any] = {}
    profile = profile or os.environ.get("AWS_PROFILE")
    if profile:
        logger.info(f"AWS Auth: Using profile: {profile}")
        boto3.setup_default_session(profile_name=profile)
    elif (
        not profile
        and os.environ.get("AWS_ACCESS_KEY_ID")
        and os.environ.get("AWS_SECRET_ACCESS_KEY")
    ):
        logger.info("AWS Auth: Using environment variables")
    else:
        raise ValueError(
            "AWS Auth: No authentication method found. "
            "Please set AWS_PROFILE (or use --profile) or AWS_ACCESS_KEY_ID and AWS_SECRET_ACCESS_KEY"
        )
    iam_client = boto3.client("iam")
    paginator = iam_client.get_paginator("get_account_authorization_details")
    for page in paginator.paginate():
        auth_report = deep_update(auth_report, page)
    a = AuthorizationDetails(auth_report)
    return a


def load_scp_from_json(inp: str) -> PolicyWithSource:
    with open(inp) as f:
        policy_description = json.load(f)
        return PolicyWithSource(
            policy_description["Policy"]["PolicySummary"]["Arn"],
            json.loads(policy_description["Policy"]["Content"]),
        )


def load_scp_from_aws(
    account_id: str, profile: Optional[str] = None
) -> List[PolicyWithSource]:
    logger.info("Attempting to fetch service control policies from AWS...")
    profile = profile or os.environ.get("AWS_PROFILE")
    policies = []

    if not profile:
        raise ValueError("No AWS profile found")

    boto3.setup_default_session(profile_name=profile)
    org_client = boto3.client("organizations")
    paginator = org_client.get_paginator("list_policies_for_target")

    for page in paginator.paginate(
        TargetId=account_id, Filter="SERVICE_CONTROL_POLICY"
    ):
        for policy in page["Policies"]:
            policy_description = org_client.describe_policy(PolicyId=policy["Id"])
            policies.append(
                PolicyWithSource(
                    policy_description["Policy"]["PolicySummary"]["Arn"],
                    json.loads(policy_description["Policy"]["Content"]),
                )
            )

    return policies


def get_auth_details(input_path: str, profile: str) -> AuthorizationDetails:
    if input_path:
        auth_details = load_auth_details_from_json(input_path)
    else:
        auth_details = load_auth_details_from_aws(profile)
    return auth_details


def get_scp_policies(
    scp_policy_arg: Optional[str], profile: Optional[str], entity_account: str
) -> List[PolicyWithSource]:
    policy_jsons = []
    if scp_policy_arg:
        scp_policy_paths = (
            scp_policy_arg.split(",")
            if "," in scp_policy_arg
            else scp_policy_arg.split(" ")
        )
        for policy_path in scp_policy_paths:
            policy_jsons.append(load_scp_from_json(inp=policy_path))

    elif profile:
        try:
            policy_jsons.extend(
                load_scp_from_aws(profile=profile, account_id=entity_account)
            )
        except ClientError as e:
            if e.response["Error"]["Code"] == "AWSOrganizationsNotInUseException":
                logger.debug("SCP Policies not in use for this account")
            elif e.response["Error"]["Code"] == "AccessDeniedException":
                logger.info(
                    "Could not fetch SCP policies due to insufficient permissions"
                )

    return policy_jsons


def build_arg_parser() -> argparse.ArgumentParser:
    arg_parser = argparse.ArgumentParser()
    basic_usage = arg_parser.add_argument_group()
    adv_usage = arg_parser.add_argument_group()
    misc_usage = arg_parser.add_argument_group()

    basic_usage.add_argument(
        "--arn",
        help="ARN of the User/Group/Role for which to evaluate the effective permissions",
    )

    adv_usage.add_argument(
        "-i",
        "--input",
        help='Path to report generated by "aws iam get-account-authorization-details"',
        required=False,
    )
    adv_usage.add_argument(
        "-s",
        "--scp-policy",
        help='Path to report(s) generated by "aws organizations describe-policy --policy-id <SCP_Policy_ID>"',
        required=False,
    )
    adv_usage.add_argument(
        "-o",
        "--output",
        help="File path to write the output to (default: stdout)",
        default="stdout",
    )
    adv_usage.add_argument(
        "-f",
        "--format",
        help="Output format",
        choices=["clean", "verbose"],
        default="clean",
    )
    adv_usage.add_argument(
        "-p",
        "--profile",
        help="AWS Profile to use when fetching authorization data",
        required=False,
    )
    adv_usage.add_argument(
        "-v", "--verbose", help="Set verbosity level to DEBUG", action="store_true"
    )

    misc_usage.add_argument(
        "-u",
        "--update",
        help="Update the IAM actions database",
        required=False,
        action="store_true",
    )

    return arg_parser


def main() -> int:
    initialize_logger()
    logger = logging.getLogger("IAM-APE")
    print_banner()

    arg_parser = build_arg_parser()
    arguments = arg_parser.parse_args()

    if arguments.verbose:
        logging.root.setLevel(logging.DEBUG)

    if not (arguments.arn or arguments.update):
        arg_parser.print_help()
        return 0

    if arguments.update:
        scrape_iam_actions()
        if not arguments.arn:
            return 0

    try:
        entity_type, entity_account = validate_arn(arguments.arn)
    except ValueError as e:
        logger.error(e)
        return -1

    try:
        auth_details = get_auth_details(arguments.input, arguments.profile)
    except (ValueError, ProfileNotFound) as e:
        logger.error(e)
        return -1

    scp_policies = None
    if arguments.input is None or arguments.scp_policy is not None:
        scp_policies = get_scp_policies(
            arguments.scp_policy, arguments.profile, entity_account
        )

    logger.info("Evaluating effective permissions")
    calculator = EffectivePolicyEvaluator(
        authorization_details=auth_details, scp_policies=scp_policies
    )

    try:
        res = calculator.evaluate(arn=arguments.arn, entity_type=entity_type)
    except ValueError as e:
        logger.error(e)
        return -1

    out: Union[AwsPolicyType, FinalReportT]
    if arguments.format == "clean":
        out = calculator.policy_expander.shrink_policy(res.allowed_permissions)
    else:
        out = calculator.create_json_report(res)

    if arguments.output == "stdout":
        logger.info(f"Effective permissions policy for {arguments.arn}\n")
        logger.info(json.dumps(out, indent=2))
    else:
        with open(arguments.output, "w") as f:
            json.dump(out, f, indent=2)
        logger.info(
            f"Wrote effective permissions for {arguments.arn} to {arguments.output}"
        )

    return 0


if __name__ == "__main__":
    exit(main())



import os
import re
from dataclasses import dataclass
from typing import Literal

RESOURCE_ARN_RE = re.compile("arn:(aws|aws-gov|aws-china):(?P<service>[^:]+):.*")


@dataclass(frozen=True)
class PolicyElement:
    WILDCARD: Literal["*"] = "*"
    EFFECT: Literal["Effect"] = "Effect"
    ALLOW: Literal["Allow"] = "Allow"
    DENY: Literal["Deny"] = "Deny"
    ACTION: Literal["Action"] = "Action"
    NOTACTION: Literal["NotAction"] = "NotAction"
    RESOURCE: Literal["Resource"] = "Resource"
    NOTRESOURCE: Literal["NotResource"] = "NotResource"
    CONDITION: Literal["Condition"] = "Condition"


CONDITIONS_NEGATIONS = {
    "StringEquals": "StringNotEquals",
    "StringNotEquals": "StringEquals",
    "StringEqualsIgnoreCase": "StringNotEqualsIgnoreCase",
    "StringNotEqualsIgnoreCase": "StringEqualsIgnoreCase",
    "StringLike": "StringNotLike",
    "StringNotLike": "StringLike",
    "NumericEquals": "NumericNotEquals",
    "NumericNotEquals": "NumericEquals",
    "NumericLessThan": "NumericGreaterThanEquals",
    "NumericGreaterThanEquals": "NumericLessThan",
    "NumericLessThanEquals": "NumericGreaterThan",
    "NumericGreaterThan": "NumericLessThanEquals",
    "DateEquals": "DateNotEquals",
    "DateNotEquals": "DateEquals",
    "DateLessThan": "DateGreaterThanEquals",
    "DateGreaterThanEquals": "DateLessThan",
    "DateLessThanEquals": "DateGreaterThan",
    "DateGreaterThan": "DateLessThanEquals",
    "IpAddress": "NotIpAddress",
    "NotIpAddress": "IpAddress",
    "ArnEquals": "ArnNotEquals",
    "ArnNotEquals": "ArnEquals",
    "ArnLike": "ArnNotLike",
    "ArnNotLike": "ArnLike",
}

actions_json_location = os.path.join(
    os.path.dirname(__file__),
    "aws_iam_actions",
    "aws_iam_actions.tar.gz",
)



import json
import logging
import os
import tarfile
import tempfile
from collections import defaultdict
from typing import Any, Dict

import requests
import tqdm  # type: ignore
from bs4 import BeautifulSoup  # type: ignore

from iam_ape.consts import actions_json_location

logger = logging.getLogger("IAM-APE:updater")
base_url = "https://docs.aws.amazon.com/service-authorization/latest/reference/"


def get_soup(url: str) -> BeautifulSoup:
    html_doc = requests.get(url).content.decode("utf-8")
    return BeautifulSoup(html_doc, "html.parser")


def scrape_iam_actions() -> int:
    soup = get_soup(base_url + "reference_policies_actions-resources-contextkeys.html")
    all_a = soup.find_all("a")
    all_links = [
        a.get("href") for a in all_a if a.get("href", "").startswith("./list_")
    ]
    data: Dict[str, Any] = defaultdict(lambda: defaultdict(dict))

    logger.info("Updating AWS IAM actions database...")
    for link in tqdm.tqdm(all_links, ncols=70):
        try:
            soup = get_soup(base_url + link[2:])
            service_prefix = soup.find("code").string
            tables = soup.find_all("div", class_="table-contents")
            for table in tables:
                headers = [
                    str(x).lower().replace("<th>", "").replace("</th>", "")
                    for x in table.find_all("th")
                ]
                if not all(
                    [header in headers for header in ("actions", "description")]
                ):
                    continue

                rowspan = 1
                for row in table.find_all("tr"):
                    if rowspan > 1:
                        rowspan -= 1
                        continue
                    rowspan = 1

                    all_cells = row.find_all("td")
                    if len(all_cells) == 0:
                        continue

                    if "rowspan" in all_cells[0].attrs:
                        rowspan = int(all_cells[0].attrs["rowspan"])

                    action = all_cells[0].text.strip().split(" ")[0]
                    description = all_cells[1].string
                    access_level = all_cells[2].string

                    data[service_prefix][action]["description"] = description
                    data[service_prefix][action]["access"] = access_level

        except AttributeError as e:
            logger.error(f"Error occurred while processing {link} - {e}")
            raise e

    logger.info("Done!")

    tmpf = os.path.join(tempfile.gettempdir(), os.urandom(24).hex())
    with open(tmpf, "w+") as f:
        json.dump(data, f, indent=2)
        f.flush()
        with tarfile.open(actions_json_location, "w:gz") as tar:
            tar.add(f.name, arcname="actions.json")
    os.remove(tmpf)

    return 0


if __name__ == "__main__":
    exit(scrape_iam_actions())



import json
import logging
import tarfile
from collections import defaultdict
from fnmatch import fnmatch
from typing import Any, Dict, FrozenSet, List, Literal, Optional, Set, Tuple

from requests.structures import CaseInsensitiveDict

from iam_ape.consts import RESOURCE_ARN_RE, PolicyElement, actions_json_location
from iam_ape.helper_classes import (
    Action,
    HashableDict,
    HashableList,
    PermissionsContainer,
    PolicyWithSource,
)
from iam_ape.helper_functions import as_list, normalize_policy
from iam_ape.helper_types import AwsPolicyStatementType, AwsPolicyType

logger = logging.getLogger("policy expander")


def _append_action(
    res: Dict[str, Set[Action]],
    action: str,
    service: str,
    resources: Optional[List[str]],
    not_resources: Optional[List[str]],
    condition: Optional[Dict[str, Any]],
    source: str,
) -> None:
    def relevant_resource(resource: str, action_service: str) -> bool:
        if resource.lower() in (
            "*",
            "arn:*",
            "arn:aws:*",
            "arn:aws-cn:*",
            "arn:aws-us-gov:*",
        ):
            return True
        if match := RESOURCE_ARN_RE.match(resource):
            resource_service = match.group("service").lower()
            if resource_service == "iam" and action_service == "sts":
                return True
            return resource_service == action_service.lower()
        return False

    for resource in resources or []:
        if not relevant_resource(resource, service):
            continue
        if condition:
            for key, val in condition.items():
                res[action].add(
                    Action(
                        action=action,
                        resource=resource,
                        not_resource=None,
                        condition=HashableDict.recursively({key: val}),
                        source=source,
                    )
                )
        else:
            res[action].add(
                Action(
                    action=action,
                    resource=resource,
                    not_resource=None,
                    condition=None,
                    source=source,
                )
            )
    for not_resource in not_resources or []:
        if not relevant_resource(not_resource, service):
            continue
        if condition:
            for key, val in condition.items():
                res[action].add(
                    Action(
                        action=action,
                        resource=None,
                        not_resource=not_resource,
                        condition=HashableDict.recursively({key: val}),
                        source=source,
                    )
                )
        else:
            res[action].add(
                Action(
                    action=action,
                    resource=None,
                    not_resource=not_resource,
                    condition=None,
                    source=source,
                )
            )


class PolicyExpander:
    def __init__(self, all_iam_actions_file_location: Optional[str] = None) -> None:
        self.all_iam_actions: CaseInsensitiveDict = self._init_iam_actions(
            all_iam_actions_file_location
        )
        self._all_service_wildcards: List[str] = [
            f"{k}:*" for k, v in self.all_iam_actions.items() if len(v) > 0
        ]

    @staticmethod
    def _init_iam_actions(
        all_iam_actions_file_location: Optional[str] = None,
    ) -> CaseInsensitiveDict:
        res: CaseInsensitiveDict = CaseInsensitiveDict()
        file_path = all_iam_actions_file_location or actions_json_location
        try:
            with tarfile.open(file_path) as f:
                data = json.load(f.extractfile("actions.json"))  # type: ignore
                for k, v in data.items():
                    res[k] = CaseInsensitiveDict(v)
        except Exception as e:
            logger.exception(f"Failed to load AWS IAM Actions due to {e}")
        return res

    def normalize_action(self, action: str) -> str:
        service, permission = action.split(":", maxsplit=1)
        permission = self.all_iam_actions[service]._store[permission.lower()][0]
        return f"{service}:{permission}"

    def expand_policies(self, policies: List[PolicyWithSource]) -> PermissionsContainer:
        allow_actions_dict: Dict[str, Set[Action]] = defaultdict(set)
        deny_actions_dict: Dict[str, Set[Action]] = defaultdict(set)

        all_actions: Dict[Literal["Allow", "Deny"], Dict[str, Set[Action]]] = {
            PolicyElement.ALLOW: allow_actions_dict,
            PolicyElement.DENY: deny_actions_dict,
        }

        for source, policy in policies:
            if policy is None:
                continue
            for i, statement in enumerate(policy["Statement"]):
                sid = "{}/{}".format(source, statement.get("Sid", i))
                if actions := statement.get(PolicyElement.ACTION):
                    for action in actions:
                        expanded_action = self.expand_action(
                            Action(
                                action=action,
                                resource=statement.get(PolicyElement.RESOURCE),
                                not_resource=statement.get(PolicyElement.NOTRESOURCE),
                                condition=statement.get(PolicyElement.CONDITION),
                                source=sid,
                            )
                        )
                        for action_key, action_values in expanded_action.items():
                            all_actions[statement[PolicyElement.EFFECT]][
                                action_key
                            ].update(action_values)

                elif statement.get(PolicyElement.NOTACTION):
                    expanded_notaction = self.expand_not_action(statement, sid)
                    for action_key, action_values in expanded_notaction.items():
                        all_actions[statement[PolicyElement.EFFECT]][action_key].update(
                            action_values
                        )

        return PermissionsContainer(
            allowed_permissions=all_actions[PolicyElement.ALLOW],
            denied_permissions=all_actions[PolicyElement.DENY],
        )

    def expand_action(self, iam_action: Action) -> Dict[str, Set[Action]]:
        res: Dict[str, Set[Action]] = defaultdict(set)
        try:
            if iam_action.action == PolicyElement.WILDCARD:  # {"Action": ["*"]}
                for service, action_dicts in self.all_iam_actions.items():
                    for action in action_dicts.keys():
                        _append_action(
                            res=res,
                            service=service,
                            action=f"{service}:{action}",
                            resources=as_list(iam_action.resource),
                            not_resources=as_list(iam_action.not_resource),
                            condition=iam_action.condition,
                            source=iam_action.source,
                        )
            elif PolicyElement.WILDCARD in iam_action.action:  # {"Action": ["iam:*"]}
                service, wildcard_action = iam_action.action.split(":", maxsplit=1)
                for action in self.all_iam_actions[service].keys():
                    if fnmatch(action.lower(), wildcard_action.lower()):
                        _append_action(
                            res=res,
                            service=service,
                            action=f"{service}:{action}",
                            resources=as_list(iam_action.resource),
                            not_resources=as_list(iam_action.not_resource),
                            condition=iam_action.condition,
                            source=iam_action.source,
                        )
            else:  # {"Action": ["sts:GetCallerIdentity"]}
                service, action = iam_action.action.split(":", maxsplit=1)
                _append_action(
                    res=res,
                    service=service,
                    action=self.normalize_action(iam_action.action),
                    resources=as_list(iam_action.resource),
                    not_resources=as_list(iam_action.not_resource),
                    condition=iam_action.condition,
                    source=iam_action.source,
                )
        except KeyError:  # not a valid action
            logger.debug(f"Got an invalid action: {iam_action.action}")

        return res

    def expand_not_action(
        self, statement: AwsPolicyStatementType, sid: str
    ) -> Dict[str, Set[Action]]:
        res: Dict[str, Set[Action]] = defaultdict(set)
        notactions: List[str] = statement.get(PolicyElement.NOTACTION) or []
        if any(
            [notaction == PolicyElement.WILDCARD for notaction in notactions]
        ):  # {"NotAction": ["*"]}
            # This is here as a safeguard. No sane person should write a policy like this. It has no effect.
            pass
        else:  # {"NotAction": ["ec2:*", "iam:Get*", "sts:GetCallerIdentity"]}
            for iam_service, action_dicts in self.all_iam_actions.items():
                for action in action_dicts.keys():
                    curr_action = f"{iam_service}:{action}"
                    curr_action_lower = curr_action.lower()
                    if any(
                        [
                            fnmatch(curr_action_lower, not_action.lower())
                            for not_action in notactions
                        ]
                    ):
                        continue
                    _append_action(
                        res=res,
                        action=curr_action,
                        service=iam_service,
                        resources=statement.get(PolicyElement.RESOURCE),
                        not_resources=statement.get(PolicyElement.NOTRESOURCE),
                        condition=statement.get(PolicyElement.CONDITION),
                        source=sid,
                    )

        return res

    def get_action_access_level(self, action: str) -> str:
        service, action_key = action.split(":", maxsplit=1)
        return self.all_iam_actions[service][action_key]["access"]

    def deflate_policy_statements(
        self,
        policy_statements: List[AwsPolicyStatementType],
    ) -> List[AwsPolicyStatementType]:
        for statement in policy_statements:
            action_dict = defaultdict(list)
            for action in statement.get(PolicyElement.ACTION) or []:
                service, action_key = action.split(":", maxsplit=1)
                action_dict[service].append(action_key)
            for service, action_keys in action_dict.items():
                try:
                    if all(
                        [
                            action in action_keys
                            for action in self.all_iam_actions[service]
                        ]
                    ):
                        for action_key in action_keys:
                            statement[PolicyElement.ACTION].remove(f"{service}:{action_key}")  # type: ignore
                        statement[PolicyElement.ACTION].append(f"{service}:*")  # type: ignore
                except KeyError as e:
                    logger.exception(f"Unknown service: {service=}")
                    raise e
            if all(
                [
                    action in as_list(statement[PolicyElement.ACTION])
                    for action in self._all_service_wildcards
                ]
            ):
                statement[PolicyElement.ACTION] = ["*"]
        return policy_statements

    def shrink_policy(self, allow_actions: Dict[str, Set[Action]]) -> AwsPolicyType:
        policy_res: AwsPolicyType = {"Statement": []}
        squashed_policies: Set[Action] = set()
        for allow_action_set in allow_actions.values():
            squashed_policies.update(allow_action_set)

        # Arrange by Resource/NotResource
        by_resource_notresource: Dict[
            Tuple[Optional[str], Optional[str]], Dict[Optional[HashableDict], Set[str]]
        ] = defaultdict(lambda: defaultdict(set))
        for action_tuple in squashed_policies:
            by_resource_notresource[(action_tuple.resource, action_tuple.not_resource)][
                HashableDict.recursively(action_tuple.condition)
            ].add(action_tuple.action)

        # Arrange by Action/Condition/NotResource
        by_action_condition_notresource: Dict[
            Tuple[FrozenSet[str], Optional[HashableDict], Optional[str]], Set[str]
        ] = defaultdict(set)
        for resource_notresource, condition_actions in by_resource_notresource.items():
            resource, notresource = resource_notresource
            for condition, actions_set in condition_actions.items():
                key = frozenset(actions_set), condition, notresource
                if resource:
                    by_action_condition_notresource[key].add(resource)
                elif not by_action_condition_notresource.get(key):
                    by_action_condition_notresource[key] = set()

        statements = list()
        for (
            action_condition_notresource,
            resources,
        ) in by_action_condition_notresource.items():
            actions, condition, notresource = action_condition_notresource
            statement: AwsPolicyStatementType = {
                "Effect": "Allow",
                PolicyElement.ACTION: sorted(list(actions)),
            }
            if resources:
                statement[PolicyElement.RESOURCE] = sorted([r for r in resources])
            if notresource:
                statement[PolicyElement.NOTRESOURCE] = as_list(notresource)
            if condition:
                statement[PolicyElement.CONDITION] = condition
            statements.append(statement)

        deflated_statements = self.deflate_policy_statements(statements)

        final_statements: Dict[int, AwsPolicyStatementType] = {}
        for statement in deflated_statements:
            statement_key = hash(
                (
                    HashableList(statement[PolicyElement.ACTION] or []),
                    HashableDict.recursively(statement.get(PolicyElement.CONDITION)),
                )
            )
            if statement_key in final_statements.keys():
                rsrc = (
                    final_statements[statement_key].get(PolicyElement.RESOURCE) or []
                ) + (statement.get(PolicyElement.RESOURCE) or [])
                if rsrc:
                    final_statements[statement_key][PolicyElement.RESOURCE] = list(
                        set(rsrc)
                    )
                not_rsrc = (
                    final_statements[statement_key].get(PolicyElement.NOTRESOURCE) or []
                ) + (statement.get(PolicyElement.NOTRESOURCE) or [])
                if not_rsrc:
                    final_statements[statement_key][PolicyElement.NOTRESOURCE] = list(
                        set(not_rsrc)
                    )
            else:
                final_statements[statement_key] = statement

        admin_statement: AwsPolicyStatementType = {
            PolicyElement.EFFECT: PolicyElement.ALLOW,
            PolicyElement.ACTION: [PolicyElement.WILDCARD],
            PolicyElement.RESOURCE: [PolicyElement.WILDCARD],
        }
        if any(
            [statement == admin_statement for statement in final_statements.values()]
        ):
            policy_res["Statement"] = [admin_statement]
        else:
            policy_res["Statement"] = list(final_statements.values())

        return normalize_policy(policy_res)



import logging
from typing import Any, Dict, List, Literal, Optional, Set, TypeVar

from iam_ape.consts import CONDITIONS_NEGATIONS
from iam_ape.helper_classes import HashableDict, HashableList
from iam_ape.helper_types import AwsPolicyType

logger = logging.getLogger(__name__)

KeyType = TypeVar("KeyType")


def as_list(element: Optional[Any]) -> List[Any]:
    if not element:
        return []
    if isinstance(element, list):
        return element
    return [element]


def normalize_policy(policy: AwsPolicyType) -> AwsPolicyType:
    def verify_type(subject: Any, subject_type: Any, err_msg: str) -> None:
        if not isinstance(subject, subject_type):
            raise TypeError(err_msg)

    def normalize_dict(
        dict_to_norm: Dict[str, Any], fields: Optional[Set[str]] = None
    ) -> Dict[str, Any]:
        for field in fields or set(dict_to_norm.keys()):
            if not isinstance(dict_to_norm.get(field, []), list):
                dict_to_norm[field] = [dict_to_norm[field]]

        return dict_to_norm

    aws_policy_str_fields: Set[str] = {"Version", "Id", "Sid"}
    aws_policy_list_fields: Set[str] = {"Statement"}
    aws_statement_list_fields: Set[str] = {
        "Action",
        "NotAction",
        "Resource",
        "NotResource",
    }
    aws_statement_action_keys: Set[Literal["Action", "NotAction"]] = {
        "Action",
        "NotAction",
    }

    if not isinstance(policy, dict):
        raise TypeError(f"Malformed policy. Expected dict, got: {type(policy)}")

    for field in aws_policy_str_fields:
        verify_type(
            policy.get(field, ""),
            str,
            err_msg=f"Malformed Policy, expected {field} of type str",
        )

    policy = normalize_dict(policy, aws_policy_list_fields)  # type: ignore

    for statement in policy["Statement"]:
        verify_type(
            statement["Effect"],
            str,
            err_msg="Malformed Policy, expected statement Effect of type str",
        )

        # Normalize list fields
        statement = normalize_dict(statement, aws_statement_list_fields)  # type: ignore

        # Normalize Principal and NotPrincipal to look like this:
        # {"Statement":
        #   {"Principal":
        #     {"AWS": ["arn:aws:s3:::some_bucket"]}
        #   }
        # }
        # As per https://docs.aws.amazon.com/IAM/latest/UserGuide/reference_policies_elements_principal.html,
        # The "*" and {"AWS": "*"} statements in Principal are equivalent, so we convert to the latter
        for field in ("Principal", "NotPrincipal"):
            if not statement.get(field):
                continue
            if statement[field] == "*":  # type: ignore
                statement[field] = {"AWS": ["*"]}  # type: ignore
            verify_type(
                statement.get(field, {}),
                dict,
                err_msg=f"Malformed Policy, expected {field} of type dict",
            )
            statement[field] = normalize_dict(statement[field])  # type: ignore

        # Normalize Condition to look like this:
        # {
        #   Statement: {
        #     Condition: {
        #       ConditionOperator: {
        #           ConditionKey: [ConditionValue]
        #       }
        #     }
        #   }
        # }
        verify_type(
            statement.get("Condition", {}),
            dict,
            err_msg="Malformed Policy, expected Condition of type dict",
        )
        for condition_operator, condition_dict in statement.get("Condition", {}).items():  # type: ignore
            for condition_key, condition_value in condition_dict.items():
                if not isinstance(condition_value, list):
                    statement["Condition"][condition_operator][condition_key] = [condition_value]  # type: ignore

        for action_key in aws_statement_action_keys:
            if action_list := statement.get(action_key):
                new_action_list = []
                for action in action_list:
                    if action.count(":") == 1:
                        service, action_name = action.split(":")
                        action = ":".join([service.lower(), action_name])
                        new_action_list.append(action)
                    else:
                        new_action_list.append(action)
                statement[action_key] = new_action_list

    return policy


def negate_condition(condition: Dict[str, Any]) -> Dict[str, Any]:
    res_condition = condition.copy()
    condition_prefix = None
    condition_key, condition_value = list(res_condition.items())[0]

    if negated := CONDITIONS_NEGATIONS.get(condition_key):
        return {negated: condition_value}

    if condition_key.count(":") == 1:
        condition_prefix, condition_key = condition_key.split(":")

    if condition_prefix == "ForAllValues":
        condition_prefix = "ForAnyValue"
    elif condition_prefix == "ForAnyValue":
        condition_prefix = "ForAllValues"
    elif condition_prefix:
        logger.warning(f"Unknown policy condition prefix: {condition_prefix}")

    if condition_key.lower().endswith("ifexists"):
        condition_key = condition_key[:-8]
        if negated := CONDITIONS_NEGATIONS.get(condition_key):
            return {f"{negated}IfExists": condition_value}

    if condition_key.lower() in ("bool", "null"):
        condition_value = {
            _condition: (
                HashableList(["true"])
                if all(v.lower() == "false" for v in _values)
                else HashableList(["false"])
                if all(v.lower() == "true" for v in _values)
                else HashableList(["true", "false"])
            )
            for _condition, _values in condition_value.items()
        }

    if condition_key == "BinaryEquals":  # there is no bloody negation
        logger.info(f"BinaryEquals used in a deny condition: {condition}")

    if condition_prefix:
        return {f"{condition_prefix}:{condition_key}": condition_value}

    return {condition_key: condition_value}


def merge_condition(
    allow_cond: Optional[Dict[str, Any]],
    deny_cond: Optional[Dict[str, Any]],
    negate: Optional[bool] = True,
) -> Optional[Dict[str, Any]]:
    res = None

    if allow_cond and not deny_cond:
        res = allow_cond

    elif negate:
        if allow_cond and deny_cond:
            res = deep_update(allow_cond, negate_condition(deny_cond))
        elif deny_cond and not allow_cond:
            res = negate_condition(deny_cond)

    else:
        if allow_cond and deny_cond:
            res = deep_update(allow_cond, deny_cond)
        else:
            res = deny_cond

    return HashableDict.recursively(res)


def get_default_policy_for_managed_policy(
    managed_policy_obj: Dict[str, Any]
) -> AwsPolicyType:
    for policy in managed_policy_obj.get("PolicyVersionList", []):
        if policy.get("IsDefaultVersion", False):
            return normalize_policy(policy["Document"])
    raise ValueError(
        f"No default policy found for managed policy {managed_policy_obj['Arn']}"
    )


def deep_update(
    mapping: Dict[KeyType, Any], *updating_mappings: Dict[KeyType, Any]
) -> Dict[KeyType, Any]:
    updated_mapping = mapping.copy()
    for updating_mapping in updating_mappings:
        for k, v in updating_mapping.items():
            if (
                k in updated_mapping
                and isinstance(updated_mapping[k], dict)
                and isinstance(v, dict)
            ):
                updated_mapping[k] = deep_update(updated_mapping[k], v)
            elif (
                k in updated_mapping
                and isinstance(updated_mapping[k], list)
                and isinstance(v, list)
            ):
                updated_mapping[k].extend(v)
            else:
                updated_mapping[k] = v
    return updated_mapping



from enum import Enum
from typing import Any, Dict, List, Optional, TypedDict


class FinalReportT(TypedDict):
    allowed_permissions: Dict[str, Any]
    denied_permissions: Dict[str, Any]
    ineffective_permissions: Dict[str, Any]


class AwsPolicyStatementType(TypedDict, total=False):
    Effect: str
    Sid: Optional[str]
    Action: Optional[List[str]]
    NotAction: Optional[List[str]]
    Resource: Optional[List[str]]
    NotResource: Optional[List[str]]
    Principal: Optional[Dict[str, Any]]
    NotPrincipal: Optional[Dict[str, Any]]
    Condition: Optional[Dict[str, Any]]


class AwsPolicyType(TypedDict, total=False):
    Version: Optional[str]
    Id: Optional[str]
    Statement: List[AwsPolicyStatementType]


class EntityType(Enum):
    user: str = "User"
    group: str = "Group"
    role: str = "Role"


class ActionDict(TypedDict):
    Action: str
    Resource: Optional[str]
    NotResource: Optional[str]
    Condition: Optional[Dict[str, Any]]
    Source: str


class IneffectiveActionDict(ActionDict):
    DeniedBy: str


class PermissionsContainerDict(TypedDict):
    allowed_permissions: List[ActionDict]
    denied_permissions: List[ActionDict]
    ineffective_permissions: List[ActionDict]



import json
import logging
from collections import defaultdict
from dataclasses import replace
from fnmatch import fnmatch
from typing import Any, Dict, Iterator, List, Optional, Set, Tuple

from iam_ape.consts import PolicyElement
from iam_ape.expand_policy import PolicyExpander
from iam_ape.helper_classes import (
    Action,
    IneffectiveAction,
    PermissionsContainer,
    PolicyWithSource,
)
from iam_ape.helper_functions import (
    deep_update,
    get_default_policy_for_managed_policy,
    merge_condition,
    normalize_policy,
)
from iam_ape.helper_types import EntityType, FinalReportT

logger = logging.getLogger("IAM-APE:evaluator")


def should_deny(
    iam_action: Action, denied_actions: Dict[str, Set[Action]]
) -> Tuple[bool, Set[Action], Optional[str]]:
    res = set()
    partially_denied = False

    for denied_action in denied_actions.get(iam_action.action, []):

        if iam_action == denied_action:
            return True, set(), denied_action.source

        if iam_action.resource:

            if denied_action.resource:
                if iam_action.resource == denied_action.resource:
                    if not denied_action.condition:
                        return True, set(), denied_action.source
                    partially_denied = True
                    res.add(
                        Action(
                            action=iam_action.action,
                            resource=iam_action.resource,
                            not_resource=None,
                            condition=merge_condition(
                                iam_action.condition, denied_action.condition
                            ),
                            source=iam_action.source,
                        )
                    )
                elif fnmatch(
                    iam_action.resource, denied_action.resource
                ):  # denied Resource > allowed Resource
                    if not denied_action.condition:
                        return True, set(), denied_action.source
                    partially_denied = True
                    res.add(
                        Action(
                            action=iam_action.action,
                            resource=iam_action.resource,
                            not_resource=None,
                            condition=merge_condition(
                                iam_action.condition, denied_action.condition
                            ),
                            source=iam_action.source,
                        )
                    )
                elif fnmatch(
                    denied_action.resource, iam_action.resource
                ):  # allowed Resource > denied Resource
                    partially_denied = True
                    res.add(
                        Action(
                            action=iam_action.action,
                            resource=None
                            if iam_action.resource == PolicyElement.WILDCARD
                            else iam_action.resource,
                            not_resource=denied_action.resource,
                            condition=merge_condition(
                                iam_action.condition, denied_action.condition
                            ),
                            source=iam_action.source,
                        )
                    )
                else:  # scopes don't overlap
                    partially_denied = True
                    res.add(
                        Action(
                            action=iam_action.action,
                            resource=None
                            if iam_action.resource == PolicyElement.WILDCARD
                            else iam_action.resource,
                            not_resource=denied_action.resource,
                            condition=merge_condition(
                                iam_action.condition, denied_action.condition
                            ),
                            source=iam_action.source,
                        )
                    )

            if denied_action.not_resource:
                if denied_action.not_resource == iam_action.resource:
                    pass
                elif fnmatch(
                    iam_action.resource, denied_action.not_resource
                ):  # denied NotResource > allowed Resource
                    pass
                elif fnmatch(
                    denied_action.not_resource, iam_action.resource
                ):  # denied NotResource < allowed Resource
                    partially_denied = True
                    res.add(
                        Action(
                            action=iam_action.action,
                            resource=denied_action.not_resource,
                            not_resource=iam_action.not_resource,
                            condition=merge_condition(
                                iam_action.condition, denied_action.condition
                            ),
                            source=iam_action.source,
                        )
                    )

        if iam_action.not_resource:

            if denied_action.resource:
                if denied_action.resource == iam_action.not_resource:
                    pass
                elif fnmatch(
                    denied_action.resource, iam_action.not_resource
                ):  # allowed NotResource > denied Resource
                    pass
                elif fnmatch(
                    iam_action.not_resource, denied_action.resource
                ):  # denied Resource > allowed NotResource
                    partially_denied = True
                    res.add(
                        Action(
                            action=iam_action.action,
                            resource=iam_action.resource,
                            not_resource=denied_action.resource,
                            condition=merge_condition(
                                iam_action.condition, denied_action.condition
                            ),
                            source=iam_action.source,
                        )
                    )
                else:  # scopes don't overlap
                    partially_denied = True
                    res.add(
                        Action(
                            action=iam_action.action,
                            resource=iam_action.resource,
                            not_resource=denied_action.resource,
                            condition=merge_condition(
                                iam_action.condition, denied_action.condition
                            ),
                            source=iam_action.source,
                        )
                    )

            if denied_action.not_resource:
                if denied_action.not_resource == iam_action.not_resource:
                    if not denied_action.condition:
                        return True, set(), denied_action.source
                    partially_denied = True
                    res.add(
                        Action(
                            action=iam_action.action,
                            resource=None,
                            not_resource=iam_action.not_resource,
                            condition=merge_condition(
                                iam_action.condition, denied_action.condition
                            ),
                            source=iam_action.source,
                        )
                    )
                elif fnmatch(
                    denied_action.not_resource, iam_action.not_resource
                ):  # allowed NotResource > denied NotResource
                    pass
                elif fnmatch(
                    iam_action.not_resource, denied_action.not_resource
                ):  # denied NotResource > allowed NotResource
                    partially_denied = True
                    res.add(
                        Action(
                            action=iam_action.action,
                            resource=iam_action.resource,
                            not_resource=denied_action.not_resource,
                            condition=merge_condition(
                                iam_action.condition, denied_action.condition
                            ),
                            source=iam_action.source,
                        )
                    )
                else:  # scopes don't overlap
                    partially_denied = True
                    res.add(
                        Action(
                            action=iam_action.action,
                            resource=denied_action.not_resource,
                            not_resource=iam_action.not_resource,
                            condition=merge_condition(
                                iam_action.condition, denied_action.condition
                            ),
                            source=iam_action.source,
                        )
                    )

    if partially_denied:
        return False, res, None
    return False, {iam_action}, None


def explicitly_deny(
    permissions: PermissionsContainer,
) -> Tuple[Dict[str, Set[Action]], Set[IneffectiveAction]]:
    final_actions_dict: Dict[str, Set[Action]] = defaultdict(set)
    ineffective_permissions: Set[IneffectiveAction] = set()
    for action_key, action_values in permissions.allowed_permissions.items():
        for action_value in action_values:
            denied, new_action_values, denied_by = should_deny(
                action_value, permissions.denied_permissions
            )
            if not denied:
                final_actions_dict[action_key].update(new_action_values)
            elif denied_by:
                ineffective_permissions.add(
                    IneffectiveAction(
                        action=action_value.action,
                        resource=action_value.resource,
                        not_resource=action_value.not_resource,
                        condition=action_value.condition,
                        source=action_value.source,
                        denied_by=denied_by,
                    )
                )

    return final_actions_dict, ineffective_permissions


def apply_permission_boundary(
    allow_actions: Dict[str, Set[Action]],
    permission_boundary: PermissionsContainer,
) -> Tuple[Dict[str, Set[Action]], Set[IneffectiveAction]]:
    def permit(at: Action, bt: Action) -> Action:
        return replace(
            at,
            condition=merge_condition(at.condition, bt.condition, negate=False),
        )

    def deny(at: Action, boundary_id: str) -> IneffectiveAction:
        return IneffectiveAction(
            action=at.action,
            resource=at.resource,
            not_resource=at.not_resource,
            condition=at.condition,
            source=at.source,
            denied_by=f"Permission Boundary: {boundary_id}",
        )

    def get_pb_id() -> str:
        if permission_boundary.allowed_permissions:
            for values in permission_boundary.allowed_permissions.values():
                for value in values:
                    return value.source
        for values in permission_boundary.denied_permissions.values():
            for value in values:
                return value.source
        raise ValueError("Permission Boundary source not found")

    permission_boundary_id = get_pb_id()

    new_allow_actions = defaultdict(set)
    ineffective_permissions: Set[IneffectiveAction] = set()
    for action_key, action_values in allow_actions.items():
        if action_key not in permission_boundary.allowed_permissions.keys():
            # action is outside of boundary
            for action_value in action_values:
                ineffective_permissions.add(deny(action_value, permission_boundary_id))
        else:
            for action_tuple in action_values:
                if action_tuple in permission_boundary.allowed_permissions[action_key]:
                    new_allow_actions[action_key].add(
                        action_tuple
                    )  # action is permitted without change
                    continue
                for boundary_tuple in permission_boundary.allowed_permissions[
                    action_key
                ]:

                    if action_tuple.resource and boundary_tuple.resource:
                        if action_tuple.resource == boundary_tuple.resource:
                            new_allow_actions[action_key].add(
                                permit(action_tuple, boundary_tuple)
                            )  # action is permitted
                        elif fnmatch(action_tuple.resource, boundary_tuple.resource):
                            new_allow_actions[action_key].add(
                                permit(action_tuple, boundary_tuple)
                            )  # action is permitted
                        elif fnmatch(boundary_tuple.resource, action_tuple.resource):
                            new_allow_actions[action_key].add(
                                replace(
                                    action_tuple,
                                    resource=boundary_tuple.resource,
                                    condition=merge_condition(
                                        action_tuple.condition,
                                        boundary_tuple.condition,
                                        negate=False,
                                    ),
                                )
                            )
                        else:  # action is not permitted
                            ineffective_permissions.add(
                                deny(action_tuple, permission_boundary_id)
                            )

                    elif action_tuple.resource and boundary_tuple.not_resource:
                        if (
                            action_tuple.resource == boundary_tuple.not_resource
                        ):  # action is not permitted
                            ineffective_permissions.add(
                                deny(action_tuple, permission_boundary_id)
                            )
                        elif fnmatch(
                            action_tuple.resource, boundary_tuple.not_resource
                        ):  # action is not permitted
                            ineffective_permissions.add(
                                deny(action_tuple, permission_boundary_id)
                            )
                        elif fnmatch(
                            boundary_tuple.not_resource, action_tuple.resource
                        ):  # action is partially permitted
                            new_allow_actions[action_key].add(
                                replace(
                                    action_tuple,
                                    not_resource=boundary_tuple.not_resource,
                                    condition=merge_condition(
                                        action_tuple.condition,
                                        boundary_tuple.condition,
                                        negate=False,
                                    ),
                                )
                            )
                        else:  # action is permitted
                            new_allow_actions[action_key].add(
                                permit(action_tuple, boundary_tuple)
                            )

                    elif action_tuple.not_resource and boundary_tuple.resource:
                        if (
                            action_tuple.not_resource == boundary_tuple.resource
                        ):  # action is not permitted
                            ineffective_permissions.add(
                                deny(action_tuple, permission_boundary_id)
                            )
                        elif fnmatch(
                            action_tuple.not_resource, boundary_tuple.resource
                        ):  # action is partially permitted
                            new_allow_actions[action_key].add(
                                replace(
                                    action_tuple,
                                    resource=boundary_tuple.resource,
                                    condition=merge_condition(
                                        action_tuple.condition,
                                        boundary_tuple.condition,
                                        negate=False,
                                    ),
                                )
                            )
                        elif fnmatch(
                            boundary_tuple.resource, action_tuple.not_resource
                        ):  # action is not permitted
                            ineffective_permissions.add(
                                deny(action_tuple, permission_boundary_id)
                            )
                        else:  # action is partially permitted
                            new_allow_actions[action_key].add(
                                replace(
                                    action_tuple,
                                    resource=boundary_tuple.resource,
                                    not_resource=None,
                                    condition=merge_condition(
                                        action_tuple.condition,
                                        boundary_tuple.condition,
                                        negate=False,
                                    ),
                                )
                            )

                    elif action_tuple.not_resource and boundary_tuple.not_resource:
                        if (
                            action_tuple.not_resource == boundary_tuple.resource
                        ):  # action is permitted
                            new_allow_actions[action_key].add(
                                permit(action_tuple, boundary_tuple)
                            )
                        elif fnmatch(
                            action_tuple.not_resource, boundary_tuple.not_resource
                        ):  # action is partially permitted
                            new_allow_actions[action_key].add(
                                replace(
                                    action_tuple,
                                    not_resource=boundary_tuple.not_resource,
                                )
                            )
                        elif fnmatch(
                            boundary_tuple.not_resource, action_tuple.not_resource
                        ):  # action is permitted
                            new_allow_actions[action_key].add(
                                permit(action_tuple, boundary_tuple)
                            )
                        else:  # action is partially permitted
                            new_allow_actions[action_key].add(
                                replace(
                                    action_tuple,
                                    not_resource=boundary_tuple.not_resource,
                                    condition=merge_condition(
                                        action_tuple.condition,
                                        boundary_tuple.condition,
                                        negate=False,
                                    ),
                                )
                            )
                            new_allow_actions[action_key].add(
                                replace(
                                    action_tuple,
                                    not_resource=action_tuple.not_resource,
                                    condition=merge_condition(
                                        action_tuple.condition,
                                        boundary_tuple.condition,
                                        negate=False,
                                    ),
                                )
                            )

    allowed, denied_ineffective = explicitly_deny(
        PermissionsContainer(
            allowed_permissions=new_allow_actions,
            denied_permissions=permission_boundary.denied_permissions,
        )
    )
    ineffective_permissions.update(denied_ineffective)
    return allowed, ineffective_permissions


class AuthorizationDetails(object):
    def __init__(self, auth_report: Optional[Dict[str, Any]] = None):
        if auth_report is None:
            raise ValueError("Could not load account authorization details")

        self.User: Dict[str, Any] = {
            user["Arn"]: user for user in auth_report.get("UserDetailList", [])
        }
        self.Group: Dict[str, Any] = {
            group["Arn"]: group for group in auth_report.get("GroupDetailList", [])
        }
        self.Role: Dict[str, Any] = {
            role["Arn"]: role for role in auth_report.get("RoleDetailList", [])
        }
        self.Policy: Dict[str, Any] = {
            policy["Arn"]: policy for policy in auth_report.get("Policies", [])
        }


class EffectivePolicyEvaluator:
    def __init__(
        self,
        authorization_details: AuthorizationDetails,
        scp_policies: Optional[List[PolicyWithSource]] = None,
        policy_expander: Optional[PolicyExpander] = None,
    ) -> None:
        self.auth_details = authorization_details
        self.policy_expander = policy_expander or PolicyExpander()
        self.scp_policy = (
            self.policy_expander.expand_policies(scp_policies)
            if scp_policies
            else PermissionsContainer()
        )

    def create_json_report(
        self, permissions_container: PermissionsContainer
    ) -> FinalReportT:
        def action_to_service(action: str) -> str:
            return action.split(":")[0]

        def serialize_set(obj):
            if isinstance(obj, set):
                return list(obj)
            return obj

        """
        {
            "allowed_permissions": {
                "<service>": {
                    "<resource>": {
                        "<access_level>": {
                            "action1": {"Condition: <condition>, "source": {<policy_arn>},
                            "action2": {"Condition: <condition>, "source": {<policy_arn>},
                        }
                    }
                }
            },
            "denied_permissions": {
                "<service>": {
                    "<resource>": {
                        "<access_level>": {
                            "action1": {"Condition: <condition>, "source": {<policy_arn>},
                            "action2": {"Condition: <condition>, "source": {<policy_arn>},
                        }
                    }
                }
            },
            "ineffective_permissions": {
                "<service>": {
                    "<resource>": {
                        "<access_level>": {
                            "action1": {
                                "denied_by": {<denied_by>}
                            },
                            "action2": {
                                "denied_by": {<denied_by>}
                            },
                        }
                    }
                }
            }
        }
        """
        res: FinalReportT = {
            "allowed_permissions": defaultdict(
                lambda: defaultdict(
                    lambda: defaultdict(
                        lambda: defaultdict(
                            lambda: {"Condition": None, "source": set()}
                        )
                    )
                )
            ),
            "denied_permissions": defaultdict(
                lambda: defaultdict(
                    lambda: defaultdict(
                        lambda: defaultdict(
                            lambda: {"Condition": None, "source": set()}
                        )
                    )
                )
            ),
            "ineffective_permissions": defaultdict(
                lambda: defaultdict(
                    lambda: defaultdict(lambda: defaultdict(lambda: defaultdict(set)))
                )
            ),
        }
        # for section in ("allowed_permissions", "denied_permissions"):
        for action_tuple_set in permissions_container.allowed_permissions.values():
            for action_tuple in action_tuple_set:
                service = action_to_service(action_tuple.action)
                resource = action_tuple.resource or "*"
                if action_tuple.not_resource:
                    if res["allowed_permissions"][service][resource].get("NotResource"):
                        res["allowed_permissions"][service][resource][
                            "NotResource"
                        ].append(action_tuple.not_resource)
                    else:
                        res["allowed_permissions"][service][resource]["NotResource"] = [
                            action_tuple.not_resource
                        ]
                access_level = self.policy_expander.get_action_access_level(
                    action_tuple.action
                )
                if cond := merge_condition(
                    res["allowed_permissions"][service][resource][access_level]
                    .get(action_tuple.action, {})
                    .get("Condition", {}),
                    action_tuple.condition,
                    negate=False,
                ):
                    res["allowed_permissions"][service][resource][access_level][
                        action_tuple.action
                    ]["Condition"] = cond
                res["allowed_permissions"][service][resource][access_level][
                    action_tuple.action
                ]["source"].add(action_tuple.source)

        for action_tuple_set in permissions_container.denied_permissions.values():
            for action_tuple in action_tuple_set:
                service = action_to_service(action_tuple.action)
                resource = action_tuple.resource or "*"
                if action_tuple.not_resource:
                    if res["denied_permissions"][service][resource].get("NotResource"):
                        res["denied_permissions"][service][resource][
                            "NotResource"
                        ].append(action_tuple.not_resource)
                    else:
                        res["denied_permissions"][service][resource]["NotResource"] = [
                            action_tuple.not_resource
                        ]
                access_level = self.policy_expander.get_action_access_level(
                    action_tuple.action
                )
                if cond := merge_condition(
                    res["denied_permissions"][service][resource][access_level]
                    .get(action_tuple.action, {})
                    .get("Condition", {}),
                    action_tuple.condition,
                    negate=False,
                ):
                    res["denied_permissions"][service][resource][access_level][
                        action_tuple.action
                    ]["Condition"] = cond

        for action_tuple in permissions_container.ineffective_permissions:
            service = action_to_service(action_tuple.action)
            resource = action_tuple.resource or "*"
            if action_tuple.not_resource:
                if res["ineffective_permissions"][service][resource].get("NotResource"):
                    res["ineffective_permissions"][service][resource][
                        "NotResource"
                    ].append(action_tuple.not_resource)
                else:
                    res["ineffective_permissions"][service][resource]["NotResource"] = [
                        action_tuple.not_resource
                    ]
            access_level = self.policy_expander.get_action_access_level(
                action_tuple.action
            )
            res["ineffective_permissions"][service][resource][access_level][
                action_tuple.action
            ]["denied_by"].add(action_tuple.denied_by)

        res = json.loads(json.dumps(res, default=serialize_set))

        return res

    def get_direct_policies(
        self, entity_obj: Dict[str, Any], entity_type: EntityType
    ) -> List[PolicyWithSource]:
        inline_policies = [
            PolicyWithSource(
                "inline policy", normalize_policy(policy["PolicyDocument"])
            )
            for policy in entity_obj.get(f"{entity_type.value}PolicyList", [])
        ]
        managed_policies = list(
            self.get_managed_policies(entity_obj.get("AttachedManagedPolicies", []))
        )
        return managed_policies + inline_policies

    def get_group_object_by_name(self, group_name: str) -> Optional[Dict[str, Any]]:
        for group in self.auth_details.Group.values():
            if group["GroupName"] == group_name:
                return group
        logger.warning(f"No such group {group_name}")
        return None

    def get_managed_policies(
        self, managed_policies: List[Dict[str, str]]
    ) -> Iterator[PolicyWithSource]:
        for policy_details in managed_policies:
            policy_arn = policy_details["PolicyArn"]
            policy = get_default_policy_for_managed_policy(
                self.auth_details.Policy.get(policy_arn, {})
            )
            yield PolicyWithSource(policy=policy, source=policy_arn)

    def get_permission_boundary(self, entity: Dict[str, Any]) -> PermissionsContainer:
        permissions = PermissionsContainer()
        if pb_arn := entity.get("PermissionsBoundary", {}).get(
            "PermissionsBoundaryArn"
        ):
            policy = PolicyWithSource(
                source=pb_arn,
                policy=get_default_policy_for_managed_policy(
                    self.auth_details.Policy.get(pb_arn, {})
                ),
            )
            permissions = self.policy_expander.expand_policies([policy])
            permissions.allowed_permissions, _ = explicitly_deny(permissions)

        return permissions

    def evaluate(self, arn: str, entity_type: EntityType) -> PermissionsContainer:
        entity_obj = getattr(self.auth_details, entity_type.value).get(arn)
        if not entity_obj:
            logger.error(f"Error - couldn't find entity with ARN {arn}")
            raise ValueError(f"couldn't find entity with ARN {arn}")

        direct_policies = self.get_direct_policies(entity_obj, entity_type)
        indirect_policies: List[PolicyWithSource] = []
        if entity_type == EntityType.user:
            for group in entity_obj.get("GroupList", []):
                group_obj = self.get_group_object_by_name(group)
                if group_obj:
                    indirect_policies.extend(
                        self.get_direct_policies(group_obj, EntityType.group)
                    )

        direct_permissions = self.policy_expander.expand_policies(
            direct_policies + indirect_policies
        )
        permission_boundary = self.get_permission_boundary(entity_obj)

        final_permissions, ineffective_permissions = explicitly_deny(direct_permissions)

        for boundary in (permission_boundary, self.scp_policy):
            if boundary.allowed_permissions or boundary.denied_permissions:
                (
                    final_permissions,
                    more_ineffective_permissions,
                ) = apply_permission_boundary(final_permissions, boundary)
                ineffective_permissions.update(more_ineffective_permissions)

            denied_permissions = deep_update(
                direct_permissions.denied_permissions,
                boundary.denied_permissions,
            )

        return PermissionsContainer(
            allowed_permissions=final_permissions,
            denied_permissions=denied_permissions,
            ineffective_permissions=ineffective_permissions,
        )
